import { existsSync, writeFileSync, readFileSync, writeFile } from 'fs';
import { promisify } from 'util';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var Detail = /** @class */ (function () {
    function Detail() {
    }
    Detail.Unescape = function (str) {
        var res = "";
        for (var i = 0; i < str.length; i++) {
            var char = str[i];
            if (char == "\\" && i != str.length - 1) {
                char = str[i + 1];
                switch (char) {
                    case "n":
                    case "\n":
                        res += "\n";
                        break;
                    case "r":
                        res += "\r";
                        break;
                    case "'":
                    case "\"":
                    case "\\":
                        res += char;
                        break;
                    default:
                        res += "\\";
                        res += char;
                        break;
                }
                continue;
            }
            res += char;
        }
        return res.trim();
    };
    Detail.Escape = function (str) {
        var res = "";
        for (var i = 0; i < str.length; i++) {
            var char = str[i];
            switch (char) {
                case "\n":
                    res += "\\n";
                    break;
                case "\r":
                    res += "\\r";
                    break;
                case "'":
                case "\"":
                case "\\":
                    res += "\\";
                    res += char;
                    break;
                default:
                    res += char;
                    break;
            }
        }
        return res;
    };
    return Detail;
}());

var NodeType;
(function (NodeType) {
    NodeType[NodeType["None"] = 0] = "None";
    NodeType[NodeType["Scalar"] = 1] = "Scalar";
    NodeType[NodeType["List"] = 2] = "List";
    NodeType[NodeType["Dict"] = 3] = "Dict";
})(NodeType || (NodeType = {}));
var Node = /** @class */ (function () {
    function Node(type, val) {
        this.type = type;
        this.value = val;
    }
    return Node;
}());

var Emitter = /** @class */ (function () {
    function Emitter() {
        this.stream = "";
    }
    Emitter.prototype.containsSpecials = function (value) {
        return /[:,'"\[\]\{\}]/gm.test(value);
    };
    Emitter.prototype.emitNode = function (node, os, indent, isLast) {
        if (indent === void 0) { indent = 0; }
        if (isLast === void 0) { isLast = true; }
        var _indent = ' '.repeat(indent * 2);
        if (node.type === NodeType.Scalar) {
            os.write("'" + Detail.Escape(node.value) + "',\n");
        }
        else if (node.type === NodeType.List) {
            os.write('[\n');
            var list = node.value;
            for (var i = 0; i < list.length; i++) {
                var it = list[i];
                os.write(_indent);
                this.emitNode(it, os, indent + 1, i == list.length - 1);
            }
            os.write("" + ' '.repeat((indent - 1) * 2) + (isLast ? ']\n' : '],\n'));
        }
        else if (node.type == NodeType.Dict) {
            if (indent > 0)
                os.write('{\n');
            var dict = node.value;
            var keys = Object.keys(dict);
            for (var i = 0; i < keys.length; i++) {
                var key = keys[i];
                if (dict[key].type == NodeType.None)
                    continue;
                os.write(_indent + key + ':');
                this.emitNode(dict[key], os, indent + 1, i == keys.length - 1);
            }
            if (indent > 0)
                os.write("" + ' '.repeat((indent - 1) * 2) + (isLast ? '}\n' : '},\n'));
        }
    };
    Emitter.prototype.emitConfigValue = function (value, indent, isLast, commas, apostrophes) {
        if (indent === void 0) { indent = 0; }
        if (isLast === void 0) { isLast = true; }
        if (commas === void 0) { commas = true; }
        if (apostrophes === void 0) { apostrophes = true; }
        var _indent = ' '.repeat(indent * 2);
        if (value instanceof Array) {
            //os.write('[\n');
            this.stream += '[\n';
            for (var i = 0; i < value.length; i++) {
                //os.write(_indent);
                this.stream += _indent;
                this.emitConfigValue(value[i], indent + 1, i == value.length - 1);
            }
            //os.write(_indent.repeat(indent - 1) + `${isLast || !commas ? ']\n' : '],\n'}`);
            this.stream += _indent.repeat(indent - 1) + ("" + (isLast || !commas ? ']\n' : '],\n'));
        }
        else if (value instanceof Object) {
            if (indent > 0)
                //os.write('{\n');
                this.stream += '{\n';
            var keys = Object.keys(value);
            for (var i = 0; i < keys.length; i++) {
                var key = keys[i];
                var _value = value[key];
                if (_value == null)
                    continue;
                //os.write(_indent + key + ':');
                this.stream += _indent + key + ':';
                this.emitConfigValue(_value, indent + 1, i == keys.length - 1);
            }
            if (indent > 0)
                //os.write(_indent.repeat(indent - 1) + `${isLast || !commas ? '}\n' : '},\n'}`);
                this.stream += _indent.repeat(indent - 1) + ("" + (isLast || !commas ? '}\n' : '},\n'));
        }
        else {
            var escaped = void 0;
            if (typeof value === "boolean") {
                escaped = Detail.Escape(String(value));
            }
            else if (typeof value === "string") {
                escaped = Detail.Escape(value);
            }
            else if (typeof value === "number") {
                escaped = Detail.Escape(value.toString());
            }
            if (escaped === undefined) {
                throw new Error("[CFG-READER] can not emit value of type: " + typeof value + ". (you passed an invalid data type)");
            }
            var useApostrophes = apostrophes || this.containsSpecials(escaped);
            this.stream += (useApostrophes ? "'" : '') + escaped + (useApostrophes ? "'" : '') + (commas ? ',' : '') + '\n';
        }
    };
    return Emitter;
}());

var TokenType;
(function (TokenType) {
    TokenType[TokenType["ArrayStart"] = 0] = "ArrayStart";
    TokenType[TokenType["ArrayEnd"] = 1] = "ArrayEnd";
    TokenType[TokenType["DictStart"] = 2] = "DictStart";
    TokenType[TokenType["DictEnd"] = 3] = "DictEnd";
    TokenType[TokenType["Key"] = 4] = "Key";
    TokenType[TokenType["Scalar"] = 5] = "Scalar";
})(TokenType || (TokenType = {}));
var ErrorType;
(function (ErrorType) {
    ErrorType[ErrorType["KeyExpected"] = 0] = "KeyExpected";
    ErrorType[ErrorType["InvalidToken"] = 1] = "InvalidToken";
    ErrorType[ErrorType["UnexpectedEOF"] = 2] = "UnexpectedEOF";
})(ErrorType || (ErrorType = {}));
var Token = /** @class */ (function () {
    function Token(_type, _value, _pos, _line, _col) {
        if (_value === void 0) { _value = ""; }
        if (_pos === void 0) { _pos = 0; }
        if (_line === void 0) { _line = 0; }
        if (_col === void 0) { _col = 0; }
        this.type = _type;
        this.value = _value;
        this.pos = _pos;
        this.line = _line;
        this.col = _col;
    }
    return Token;
}());
var Parser = /** @class */ (function () {
    function Parser(content, filePath) {
        this.tokens = [];
        this.readPos = 0;
        this.line = 0;
        this.column = 0;
        this.tokIdx = 0;
        this.buffer = content;
        this.filePath = filePath;
    }
    Parser.prototype.parse = function () {
        this.tokenize();
        return this.parseToken();
    };
    Parser.prototype.unread = function () {
        return this.buffer.length - this.readPos;
    };
    Parser.prototype.peek = function (offset) {
        if (offset === void 0) { offset = 0; }
        var idx = this.readPos + offset;
        return this.buffer[idx];
    };
    Parser.prototype.get = function () {
        this.column++;
        if (this.peek() == '\n') {
            this.line++;
            this.column = 0;
        }
        //const currPos = this.readPos;
        //this.readPos++;
        //return this.buffer[currPos];
        return this.buffer[this.readPos++];
    };
    Parser.prototype.skip = function (n) {
        if (n === void 0) { n = 1; }
        for (var i = 0; i < n; i++) {
            this.column++;
            if (this.peek(i) == '\n') {
                this.line++;
                this.column = 0;
            }
        }
        this.readPos += n;
    };
    Parser.prototype.skipNextToken = function () {
        while (this.unread() > 0) {
            if (this.peek() == ' ' || this.peek() == '\n' || this.peek() == '\r' || this.peek() == '\t' || this.peek() == ',') {
                this.skip();
            }
            else if (this.peek() == '#') {
                this.skip();
                while (this.unread() > 0 && this.peek() != '\n' && this.peek() != '#') {
                    this.skip();
                }
                if (this.unread() > 0) {
                    this.skip();
                }
            }
            else {
                break;
            }
        }
    };
    Parser.prototype.tokenize = function () {
        this.tokens.push(new Token(TokenType.DictStart));
        while (this.unread() > 0) {
            this.skipNextToken();
            if (this.unread() == 0) {
                break;
            }
            if (this.peek() == '[') {
                this.skip();
                this.tokens.push(new Token(TokenType.ArrayStart, "", this.readPos, this.line, this.column));
            }
            else if (this.peek() == ']') {
                this.skip();
                this.tokens.push(new Token(TokenType.ArrayEnd, "", this.readPos, this.line, this.column));
            }
            else if (this.peek() == '{') {
                this.skip();
                this.tokens.push(new Token(TokenType.DictStart, "", this.readPos, this.line, this.column));
            }
            else if (this.peek() == '}') {
                this.skip();
                this.tokens.push(new Token(TokenType.DictEnd, "", this.readPos, this.line, this.column));
            }
            else {
                var val = "";
                if (this.peek() == '\'' || this.peek() == '"') {
                    var start = this.get();
                    if (this.peek() != start) {
                        while (this.unread() > 1 && (this.peek() == '\\' || this.peek(1) != start)) {
                            if (this.peek() == '\n' || this.peek() == '\r') {
                                if (this.get() == '\r' || this.peek() == '\n') {
                                    this.skip();
                                }
                                val += "\n";
                                continue;
                            }
                            val += this.get();
                        }
                        if (this.unread() > 0) {
                            val += this.get();
                        }
                        if (this.unread() == 0) {
                            throw new Error(this.createParseError(ErrorType.UnexpectedEOF, this.line, this.column));
                        }
                    }
                    this.skip();
                }
                else {
                    while (this.unread() > 0 &&
                        this.peek() != '\n' &&
                        this.peek() != ':' &&
                        this.peek() != ',' &&
                        this.peek() != ']' &&
                        this.peek() != '}' &&
                        this.peek() != '#') {
                        val += this.get();
                    }
                }
                val = Detail.Unescape(val);
                if (this.unread() > 0 && this.peek() == ':') {
                    this.tokens.push(new Token(TokenType.Key, val, this.readPos, this.line, this.column));
                }
                else {
                    this.tokens.push(new Token(TokenType.Scalar, val, this.readPos, this.line, this.column));
                }
                if (this.unread() > 0 && (this.peek() == ':' || this.peek() == ',')) {
                    this.skip();
                }
            }
        }
        this.tokens.push(new Token(TokenType.DictEnd));
        return; // end
    };
    Parser.prototype.createParseError = function (type, token, col) {
        var line;
        if (token instanceof Token) {
            col = token.col;
            line = token.line;
        }
        else
            line = token;
        line++;
        col++;
        var place = this.filePath ? this.filePath + ":" + line + ":" + col : line + ":" + col;
        var base = "[CFG-READER] error at line " + place + " -> ";
        switch (type) {
            case ErrorType.KeyExpected:
                return base + "key expected";
            case ErrorType.InvalidToken:
                return base + "invalid token";
            case ErrorType.UnexpectedEOF:
                return base + "unexpected end of file";
        }
    };
    Parser.prototype.parseToken = function () {
        var token = this.tokens[this.tokIdx];
        switch (token.type) {
            case TokenType.Scalar:
                return new Node(NodeType.Scalar, token.value);
            case TokenType.ArrayStart:
                var list = new Node(NodeType.List, []);
                while (this.tokIdx < this.tokens.length - 1 && this.tokens[this.tokIdx + 1].type != TokenType.ArrayEnd) {
                    this.tokIdx++;
                    var node = this.parseToken();
                    list.value.push(node);
                }
                this.tokIdx++;
                return list;
            case TokenType.DictStart:
                var dict = new Node(NodeType.Dict, {});
                while (this.tokIdx < this.tokens.length - 1 && this.tokens[this.tokIdx + 1].type != TokenType.DictEnd) {
                    this.tokIdx++;
                    var nextTok = this.tokens[this.tokIdx];
                    if (nextTok.type != TokenType.Key) {
                        throw new Error(this.createParseError(ErrorType.KeyExpected, nextTok));
                    }
                    var key = nextTok.value;
                    this.tokIdx++;
                    var node = this.parseToken();
                    dict.value[key] = node;
                }
                this.tokIdx++;
                return dict;
        }
        throw new Error(this.createParseError(ErrorType.InvalidToken, token));
    };
    return Parser;
}());

var Config = /** @class */ (function () {
    /**
     *
     * @param {string} fileName
     * @param {Object} predefinedValues [optional]
     */
    function Config(fileName, preDefines) {
        this.config = {};
        if (typeof fileName !== "string") {
            throw new Error("[CFG-READER]: invalid constructor call, fileName must be type string");
        }
        this.fileName = fileName;
        if (!(preDefines instanceof Object) && preDefines != null) {
            throw new Error("[CFG-READER]: invalid constructor call, preDefines must be null or Object");
        }
        if (preDefines == null && this.existsFile(fileName)) {
            this.loadFile(fileName);
            this.parse();
        }
        else if (preDefines instanceof Object) {
            this.config = preDefines;
            if (this.existsFile(fileName)) {
                this.loadFile(fileName);
                this.parse();
            }
        }
    }
    Config.prototype.existsFile = function (path) {
        return existsSync(path);
    };
    Config.prototype.createFile = function (path) {
        writeFileSync(path, "", { encoding: "utf8" });
    };
    Config.prototype.loadFile = function (path) {
        this.content = readFileSync(path, { encoding: "utf8" });
    };
    Config.prototype.parseNode = function (node) {
        if (node.type == NodeType.Dict) {
            var dict = {};
            for (var key in node.value) {
                var valueNode = node.value[key];
                var value = this.parseNode(valueNode);
                dict[key] = value;
            }
            return dict;
        }
        else if (node.type == NodeType.List) {
            var length_1 = node.value.length;
            var list = new Array(length_1);
            for (var i = 0; i < length_1; i++) {
                var valueNode = node.value[i];
                var value = this.parseNode(valueNode);
                list[i] = value;
            }
            return list;
        }
        else if (node.type == NodeType.Scalar) {
            var value = node.value;
            if (value === "true" ||
                value === "false" ||
                value === "yes" ||
                value === "no") {
                return value === "true" || value === "yes";
            }
            else if (isNaN(parseFloat(value))) {
                return value;
            }
            else {
                return parseFloat(value);
            }
        }
        return null;
    };
    Config.prototype.parse = function () {
        if (this.content == null) {
            throw new Error("[CFG-READER]: no file loaded (internal)");
        }
        this.parser = new Parser(this.content, this.fileName);
        var node = this.parser.parse();
        var config = this.parseNode(node);
        this.config = Object.assign(this.config, config);
    };
    /**
     * Get a config value with unknown type, slower than GetOfType
     * @param {string} key
     * @returns {ConfigValue}
     */
    Config.prototype.get = function (key) {
        return this.config[key];
    };
    /**
     * Set a config value
     * @param {string} key
     * @param {ConfigValue} value
     */
    Config.prototype.set = function (key, value) {
        this.config[key] = value;
    };
    /**
     * Save the current changes to the opened file
     * @param {boolean} useCommas [default: true]
     * @param {boolean} useApostrophe [default: true]
     *
     * @returns {Promise<void>}
     */
    Config.prototype.save = function (useCommas, useApostrophe) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!this.existsFile(this.fileName))
                            this.createFile(this.fileName);
                        this.emitter = new Emitter();
                        this.emitter.emitConfigValue(this.config, 0, true, useCommas, useApostrophe);
                        return [4 /*yield*/, promisify(writeFile)(this.fileName, this.emitter.stream, {
                                encoding: "utf8",
                            })];
                    case 1:
                        _a.sent();
                        return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Get a config value with known type, faster than normal Get
     * @param {string} key
     * @param {ValueType} type
     * @returns {ReturnValueType}
     */
    Config.prototype.getOfType = function (key) {
        return this.config[key];
    };
    /**
     * Serialize config
     * @param {boolean} useCommas [default: true]
     * @param {boolean} useApostrophe [default: true]
     * @returns {string}
     */
    Config.prototype.serialize = function (useCommas, useApostrophe) {
        this.emitter = new Emitter();
        this.emitter.emitConfigValue(this.config, 0, true, useCommas, useApostrophe);
        return this.emitter.stream;
    };
    return Config;
}());

export { Config };
